@@ -63,16 +63,8 @@
 GST_DEBUG_CATEGORY_STATIC (gst_gl_upload_debug);
 #define GST_CAT_DEFAULT gst_gl_upload_debug
 
-#define DEBUG_INIT \
-  GST_DEBUG_CATEGORY_INIT (gst_gl_upload_debug, "glupload", 0, "upload");
-
-G_DEFINE_TYPE_WITH_CODE (GstGLUpload, gst_gl_upload, GST_TYPE_OBJECT,
-    DEBUG_INIT);
 static void gst_gl_upload_finalize (GObject * object);
 
-#define GST_GL_UPLOAD_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
-    GST_TYPE_GL_UPLOAD, GstGLUploadPrivate))
-
 static GstGLTextureTarget
 _caps_get_texture_target (GstCaps * caps, GstGLTextureTarget default_target)
 {
@@ -111,8 +103,17 @@ struct _GstGLUploadPrivate
   const UploadMethod *method;
   gpointer method_impl;
   int method_i;
+
+  /* saved method for reconfigure */
+  int saved_method_i;
 };
 
+#define DEBUG_INIT \
+  GST_DEBUG_CATEGORY_INIT (gst_gl_upload_debug, "glupload", 0, "upload");
+
+G_DEFINE_TYPE_WITH_CODE (GstGLUpload, gst_gl_upload, GST_TYPE_OBJECT,
+    G_ADD_PRIVATE (GstGLUpload) DEBUG_INIT);
+
 static GstCaps *
 _set_caps_features_with_passthrough (const GstCaps * caps,
     const gchar * feature_name, GstCapsFeatures * passthrough)
@@ -490,6 +491,12 @@ struct DmabufUpload
   GstGLFormat formats[GST_VIDEO_MAX_PLANES];
   GstBuffer *outbuf;
   GstGLVideoAllocationParams *params;
+  guint n_mem;
+
+  gboolean direct;
+  GstVideoInfo out_info;
+  /* only used for pointer comparision */
+  gpointer out_caps;
 };
 
 static GstStaticCaps _dma_buf_upload_caps =
@@ -510,11 +517,22 @@ static GstCaps *
 _dma_buf_upload_transform_caps (gpointer impl, GstGLContext * context,
     GstPadDirection direction, GstCaps * caps)
 {
-  GstCapsFeatures *passthrough =
-      gst_caps_features_from_string
-      (GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION);
+  struct DmabufUpload *dmabuf = impl;
+  GstCapsFeatures *passthrough;
   GstCaps *ret;
 
+  if (context) {
+    /* Don't propose DMABuf caps feature unless it can be supported */
+    if (gst_gl_context_get_gl_platform (context) != GST_GL_PLATFORM_EGL)
+      return NULL;
+
+    if (!gst_gl_context_check_feature (context, "EGL_KHR_image_base"))
+      return NULL;
+  }
+
+  passthrough = gst_caps_features_from_string
+      (GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION);
+
   if (direction == GST_PAD_SINK) {
     GstCaps *tmp;
 
@@ -527,10 +545,16 @@ _dma_buf_upload_transform_caps (gpointer impl, GstGLContext * context,
     ret = tmp;
   } else {
     gint i, n;
+    GstCaps *tmp;
 
     ret =
         _set_caps_features_with_passthrough (caps,
         GST_CAPS_FEATURE_MEMORY_DMABUF, passthrough);
+    tmp =
+        _set_caps_features_with_passthrough (caps,
+        GST_CAPS_FEATURE_MEMORY_SYSTEM_MEMORY, passthrough);
+    gst_caps_append (ret, tmp);
+
 
     n = gst_caps_get_size (ret);
     for (i = 0; i < n; i++) {
@@ -542,18 +566,22 @@ _dma_buf_upload_transform_caps (gpointer impl, GstGLContext * context,
 
   gst_caps_features_free (passthrough);
 
+  GST_DEBUG_OBJECT (dmabuf->upload, "transformed %" GST_PTR_FORMAT " into %"
+      GST_PTR_FORMAT, caps, ret);
+
   return ret;
 }
 
 static GQuark
 _eglimage_quark (gint plane)
 {
-  static GQuark quark[4] = { 0 };
+  static GQuark quark[5] = { 0 };
   static const gchar *quark_str[] = {
     "GstGLDMABufEGLImage0",
     "GstGLDMABufEGLImage1",
     "GstGLDMABufEGLImage2",
     "GstGLDMABufEGLImage3",
+    "GstGLDMABufEGLImage",
   };
 
   if (!quark[plane])
@@ -582,12 +610,13 @@ _dma_buf_upload_accept (gpointer impl, GstBuffer * buffer, GstCaps * in_caps,
 {
   struct DmabufUpload *dmabuf = impl;
   GstVideoInfo *in_info = &dmabuf->upload->priv->in_info;
+  GstVideoInfo *out_info = &dmabuf->out_info;
   guint n_planes = GST_VIDEO_INFO_N_PLANES (in_info);
   GstVideoMeta *meta;
   guint n_mem;
-  guint mems_idx[GST_VIDEO_MAX_PLANES];
-  gsize mems_skip[GST_VIDEO_MAX_PLANES];
   GstMemory *mems[GST_VIDEO_MAX_PLANES];
+  gsize offset[GST_VIDEO_MAX_PLANES];
+  gint fd[GST_VIDEO_MAX_PLANES];
   guint i;
 
   n_mem = gst_buffer_n_memory (buffer);
@@ -621,11 +650,17 @@ _dma_buf_upload_accept (gpointer impl, GstBuffer * buffer, GstCaps * in_caps,
     }
   }
 
+  if (out_caps != dmabuf->out_caps) {
+    dmabuf->out_caps = out_caps;
+    if (!gst_video_info_from_caps (out_info, out_caps))
+      return FALSE;
+  }
+
   if (dmabuf->params)
     gst_gl_allocation_params_free ((GstGLAllocationParams *) dmabuf->params);
   if (!(dmabuf->params =
           gst_gl_video_allocation_params_new_wrapped_gl_handle (dmabuf->
-              upload->context, NULL, &dmabuf->upload->priv->in_info, -1, NULL,
+              upload->context, NULL, out_info, -1, NULL,
               GST_GL_TEXTURE_TARGET_2D, 0, NULL, NULL, NULL)))
     return FALSE;
 
@@ -633,41 +668,56 @@ _dma_buf_upload_accept (gpointer impl, GstBuffer * buffer, GstCaps * in_caps,
   for (i = 0; i < n_planes; i++) {
     guint plane_size;
     guint length;
+    guint mem_idx;
+    gsize mem_skip;
 
     plane_size = gst_gl_get_plane_data_size (in_info, NULL, i);
 
     if (!gst_buffer_find_memory (buffer, in_info->offset[i], plane_size,
-            &mems_idx[i], &length, &mems_skip[i]))
+            &mem_idx, &length, &mem_skip))
       return FALSE;
 
     /* We can't have more then one dmabuf per plane */
     if (length != 1)
       return FALSE;
 
-    mems[i] = gst_buffer_peek_memory (buffer, mems_idx[i]);
+    mems[i] = gst_buffer_peek_memory (buffer, mem_idx);
 
     /* And all memory found must be dmabuf */
     if (!gst_is_dmabuf_memory (mems[i]))
       return FALSE;
+
+    offset[i] = mems[i]->offset + mem_skip;
+    fd[i] = gst_dmabuf_memory_get_fd (mems[i]);
   }
 
+  if (dmabuf->direct)
+    dmabuf->n_mem = 1;
+  else
+    dmabuf->n_mem = n_planes;
+
   /* Now create an EGLImage for each dmabufs */
-  for (i = 0; i < n_planes; i++) {
+  for (i = 0; i < dmabuf->n_mem; i++) {
+    gint cache_id = dmabuf->direct ? 4 : i;
+
     /* check if one is cached */
-    dmabuf->eglimage[i] = _get_cached_eglimage (mems[i], i);
+    dmabuf->eglimage[i] = _get_cached_eglimage (mems[i], cache_id);
     if (dmabuf->eglimage[i])
       continue;
 
     /* otherwise create one and cache it */
-    dmabuf->eglimage[i] =
-        gst_egl_image_from_dmabuf (dmabuf->upload->context,
-        gst_dmabuf_memory_get_fd (mems[i]), in_info, i,
-        mems[i]->offset + mems_skip[i]);
+    if (dmabuf->direct)
+      dmabuf->eglimage[i] =
+          gst_egl_image_from_dmabuf_direct (dmabuf->upload->context, fd, offset,
+          in_info);
+    else
+      dmabuf->eglimage[i] = gst_egl_image_from_dmabuf (dmabuf->upload->context,
+          fd[i], in_info, i, offset[i]);
 
     if (!dmabuf->eglimage[i])
       return FALSE;
 
-    _set_cached_eglimage (mems[i], dmabuf->eglimage[i], i);
+    _set_cached_eglimage (mems[i], dmabuf->eglimage[i], cache_id);
     dmabuf->formats[i] = dmabuf->eglimage[i]->format;
   }
 
@@ -694,8 +744,7 @@ _dma_buf_upload_perform_gl_thread (GstGLContext * context,
   /* FIXME: buffer pool */
   dmabuf->outbuf = gst_buffer_new ();
   gst_gl_memory_setup_buffer (allocator, dmabuf->outbuf, dmabuf->params,
-      dmabuf->formats, (gpointer *) dmabuf->eglimage,
-      gst_buffer_n_memory (dmabuf->outbuf));
+      dmabuf->formats, (gpointer *) dmabuf->eglimage, dmabuf->n_mem);
   gst_object_unref (allocator);
 }
 
@@ -741,6 +790,98 @@ static const UploadMethod _dma_buf_upload = {
   &_dma_buf_upload_free
 };
 
+/* a variant of the DMABuf uploader that relies on HW color convertion instead
+ * of shaders */
+
+static gpointer
+_direct_dma_buf_upload_new (GstGLUpload * upload)
+{
+  struct DmabufUpload *dmabuf = _dma_buf_upload_new (upload);
+  dmabuf->direct = TRUE;
+  gst_video_info_init (&dmabuf->out_info);
+  return dmabuf;
+}
+
+static GstCaps *
+_direct_dma_buf_upload_transform_caps (gpointer impl, GstGLContext * context,
+    GstPadDirection direction, GstCaps * caps)
+{
+  struct DmabufUpload *dmabuf = impl;
+  GstCapsFeatures *passthrough =
+      gst_caps_features_from_string
+      (GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION);
+  GstCaps *ret;
+
+  if (direction == GST_PAD_SINK) {
+    gint i, n;
+    GstCaps *tmp;
+
+    ret =
+        _set_caps_features_with_passthrough (caps,
+        GST_CAPS_FEATURE_MEMORY_GL_MEMORY, passthrough);
+
+    gst_caps_set_simple (ret, "format", G_TYPE_STRING, "RGBA", NULL);
+
+    n = gst_caps_get_size (ret);
+    for (i = 0; i < n; i++) {
+      GstStructure *s = gst_caps_get_structure (ret, i);
+
+      gst_structure_remove_fields (s, "chroma-site", NULL);
+      gst_structure_remove_fields (s, "colorimetry", NULL);
+    }
+    tmp = _caps_intersect_texture_target (ret, 1 << GST_GL_TEXTURE_TARGET_2D);
+    gst_caps_unref (ret);
+    ret = tmp;
+  } else {
+    gint i, n;
+    GstCaps *tmp;
+    GValue formats = G_VALUE_INIT;
+    gchar *format_str = g_strdup (GST_GL_MEMORY_VIDEO_FORMATS_STR);
+
+    ret =
+        _set_caps_features_with_passthrough (caps,
+        GST_CAPS_FEATURE_MEMORY_DMABUF, passthrough);
+    tmp =
+        _set_caps_features_with_passthrough (caps,
+        GST_CAPS_FEATURE_MEMORY_SYSTEM_MEMORY, passthrough);
+    gst_caps_append (ret, tmp);
+
+    g_value_init (&formats, GST_TYPE_LIST);
+    gst_value_deserialize (&formats, format_str);
+    tmp = gst_caps_copy (ret);
+    gst_caps_set_value (tmp, "format", &formats);
+    gst_caps_append (ret, tmp);
+    g_free (format_str);
+    g_value_unset (&formats);
+
+    n = gst_caps_get_size (ret);
+    for (i = 0; i < n; i++) {
+      GstStructure *s = gst_caps_get_structure (ret, i);
+
+      gst_structure_remove_fields (s, "texture-target", NULL);
+    }
+  }
+
+  gst_caps_features_free (passthrough);
+
+  GST_DEBUG_OBJECT (dmabuf->upload, "transformed %" GST_PTR_FORMAT " into %"
+      GST_PTR_FORMAT, caps, ret);
+
+  return ret;
+}
+
+static const UploadMethod _direct_dma_buf_upload = {
+  "DirectDmabuf",
+  0,
+  &_dma_buf_upload_caps,
+  &_direct_dma_buf_upload_new,
+  &_direct_dma_buf_upload_transform_caps,
+  &_dma_buf_upload_accept,
+  &_dma_buf_upload_propose_allocation,
+  &_dma_buf_upload_perform,
+  &_dma_buf_upload_free
+};
+
 #endif /* GST_GL_HAVE_DMABUF */
 
 struct GLUploadMeta
@@ -1527,6 +1668,7 @@ static const UploadMethod _directviv_upload = {
 
 static const UploadMethod *upload_methods[] = { &_gl_memory_upload,
 #if GST_GL_HAVE_DMABUF
+  &_direct_dma_buf_upload,
   &_dma_buf_upload,
 #endif
 #if GST_GL_HAVE_VIV_DIRECTVIV
@@ -1562,15 +1704,13 @@ gst_gl_upload_get_input_template_caps (void)
 static void
 gst_gl_upload_class_init (GstGLUploadClass * klass)
 {
-  g_type_class_add_private (klass, sizeof (GstGLUploadPrivate));
-
   G_OBJECT_CLASS (klass)->finalize = gst_gl_upload_finalize;
 }
 
 static void
 gst_gl_upload_init (GstGLUpload * upload)
 {
-  upload->priv = GST_GL_UPLOAD_GET_PRIVATE (upload);
+  upload->priv = gst_gl_upload_get_instance_private (upload);
 }
 
 /**
@@ -1654,6 +1794,21 @@ gst_gl_upload_transform_caps (GstGLUpload * upload, GstGLContext * context,
   GstCaps *result, *tmp;
   gint i;
 
+  if (upload->priv->method) {
+    tmp = upload->priv->method->transform_caps (upload->priv->method_impl,
+        context, direction, caps);
+    if (filter) {
+      result = gst_caps_intersect_full (filter, tmp, GST_CAPS_INTERSECT_FIRST);
+      gst_caps_unref (tmp);
+    } else {
+      result = tmp;
+    }
+    if (!gst_caps_is_empty (result))
+      return result;
+    else
+      gst_caps_unref (result);
+  }
+
   tmp = gst_caps_new_empty ();
 
   for (i = 0; i < G_N_ELEMENTS (upload_methods); i++) {
@@ -1714,7 +1869,6 @@ _gst_gl_upload_set_caps_unlocked (GstGLUpload * upload, GstCaps * in_caps,
   gst_video_info_from_caps (&upload->priv->in_info, in_caps);
   gst_video_info_from_caps (&upload->priv->out_info, out_caps);
 
-  upload->priv->method = NULL;
   upload->priv->method_impl = NULL;
   upload->priv->method_i = 0;
 
@@ -1765,14 +1919,29 @@ gst_gl_upload_get_caps (GstGLUpload * upload, GstCaps ** in_caps,
 }
 
 static gboolean
-_upload_find_method (GstGLUpload * upload)
+_upload_find_method (GstGLUpload * upload, gpointer last_impl)
 {
   gint method_i;
 
-  if (upload->priv->method_i >= G_N_ELEMENTS (upload_methods))
-    return FALSE;
+  /* start with the last used method after explicitly reconfiguring to
+   * negotiate caps for this method */
+  if (upload->priv->method_i == 0) {
+    upload->priv->method_i = upload->priv->saved_method_i;
+    upload->priv->saved_method_i = 0;
+  }
+
+  if (upload->priv->method_i >= G_N_ELEMENTS (upload_methods)) {
+    if (last_impl)
+      upload->priv->method_i = 0;
+    else
+      return FALSE;
+  }
 
   method_i = upload->priv->method_i;
+
+  if (last_impl == upload->priv->upload_impl[method_i])
+    return FALSE;
+
   upload->priv->method = upload_methods[method_i];
   upload->priv->method_impl = upload->priv->upload_impl[method_i];
 
@@ -1801,6 +1970,7 @@ gst_gl_upload_perform_with_buffer (GstGLUpload * upload, GstBuffer * buffer,
 {
   GstGLUploadReturn ret = GST_GL_UPLOAD_ERROR;
   GstBuffer *outbuf;
+  gpointer last_impl = upload->priv->method_impl;
 
   g_return_val_if_fail (GST_IS_GL_UPLOAD (upload), FALSE);
   g_return_val_if_fail (GST_IS_BUFFER (buffer), FALSE);
@@ -1810,7 +1980,7 @@ gst_gl_upload_perform_with_buffer (GstGLUpload * upload, GstBuffer * buffer,
 
 #define NEXT_METHOD \
 do { \
-  if (!_upload_find_method (upload)) { \
+  if (!_upload_find_method (upload, last_impl)) { \
     GST_OBJECT_UNLOCK (upload); \
     return FALSE; \
   } \
@@ -1818,7 +1988,7 @@ do { \
 } while (0)
 
   if (!upload->priv->method_impl)
-    _upload_find_method (upload);
+    _upload_find_method (upload, last_impl);
 
 restart:
   if (!upload->priv->method->accept (upload->priv->method_impl, buffer,
@@ -1842,6 +2012,15 @@ restart:
     }
     goto restart;
   } else if (ret == GST_GL_UPLOAD_DONE || ret == GST_GL_UPLOAD_RECONFIGURE) {
+    if (last_impl != upload->priv->method_impl) {
+      GstCaps *caps = gst_gl_upload_transform_caps (upload, upload->context,
+          GST_PAD_SINK, upload->priv->in_caps, NULL);
+      if (!gst_caps_is_subset (caps, upload->priv->out_caps)) {
+        gst_buffer_replace (&outbuf, NULL);
+        ret = GST_GL_UPLOAD_RECONFIGURE;
+      }
+      gst_caps_unref (caps);
+    }
     /* we are done */
   } else {
     upload->priv->method_impl = NULL;
@@ -1853,6 +2032,9 @@ restart:
         GST_BUFFER_COPY_FLAGS | GST_BUFFER_COPY_TIMESTAMPS, 0, -1);
   *outbuf_ptr = outbuf;
 
+  if (ret == GST_GL_UPLOAD_RECONFIGURE)
+    upload->priv->saved_method_i = upload->priv->method_i - 1;
+
   GST_OBJECT_UNLOCK (upload);
 
   return ret;
